open Base

type t = Cookie.t list [@@deriving show]

let empty = []

let of_list =
  let compare c1 c2 = String.compare (Cookie.get_key c1) (Cookie.get_key c2) in
  List.dedup_and_sort ~compare

let of_cookie_pair p =
  let ws_or_dquote c = Char.(is_whitespace c || equal c '"') in
  String.lsplit2 ~on:'=' p
  |> Option.map ~f:begin fun (k, v) ->
    Cookie.make (String.strip k, Uri.pct_decode (String.strip ~drop:ws_or_dquote v))
  end

let of_cookie_header str =
  str
  |> String.split ~on:';'
  |> List.filter_map ~f:of_cookie_pair
  |> of_list

let imf_fixdate_fmt =
  "%a, %d %b %Y %H:%M:%S GMT"

let gmt_zone =
  Core.Time.Zone.of_string "GMT"

let of_set_cookie_headers headers =
  let of_imf_fixdate str =
    Option.try_with begin fun () ->
      Core.Time.parse str ~fmt:imf_fixdate_fmt ~zone:gmt_zone
    end in
  let int64_of_str str =
    Option.try_with begin fun () ->
      Int64.of_string str
    end in
  let set_attr cookie attr = match String.strip attr with
    | "Secure" ->
      Cookie.set_secure true cookie
    | "HttpOnly" ->
      Cookie.set_http_only true cookie
    | _ ->
      String.lsplit2 ~on:'=' attr
      |> Option.map ~f:begin fun (k, v) -> match String.strip k with
        | "Expires" ->
          let expires = of_imf_fixdate (String.strip v) in
          Cookie.set_expires expires cookie
        | "MaxAge" ->
          let max_age = int64_of_str (String.strip v) in
          Cookie.set_max_age max_age cookie
        | "Domain" ->
          Cookie.set_domain (Some (String.strip v)) cookie
        | "Path" ->
          Cookie.set_path (Some (String.strip v)) cookie
        | "SameSite" ->
          let same_site = match String.strip v with
            | "Strict" -> Some Cookie.Same_site.Strict
            | "Lax" -> Some Cookie.Same_site.Lax
            | "None" -> Some Cookie.Same_site.None
            | _ -> None
          in
          Cookie.set_same_site same_site cookie
        | _ -> cookie
      end
      |> Option.value ~default:cookie in
  let of_set_cookie_header hdr =
    match String.split ~on:';' hdr with
    | pair :: attrs ->
      of_cookie_pair pair
      |> Option.map ~f:(fun init -> List.fold attrs ~init ~f:set_attr)
    | [] -> None
  in
  List.filter_map headers ~f:of_set_cookie_header |> of_list

let to_list =
  Fn.id

let to_cookie_pair c =
  let k = Cookie.get_key c in
  let v = (c |> Cookie.get_value |> Uri.pct_encode ~component:`Query) in
  Printf.sprintf "%s=%s" k v

let to_cookie_header cookies =
  cookies
  |> List.map ~f:to_cookie_pair
  |> String.concat ~sep:";"

let to_set_cookie_headers cookies =
  let open Printf in
  let to_imf_fixdate time =
    Core.Time.format time imf_fixdate_fmt ~zone:gmt_zone in
  let to_set_cookie_header c =
    let cookie_pair = to_cookie_pair c in
    let default = "" in
    let expires =
      Cookie.get_expires c
      |> Option.value_map ~f:(fun v -> sprintf ";Expires=%s" (to_imf_fixdate v)) ~default in
    let max_age =
      Cookie.get_max_age c
      |> Option.value_map ~f:(fun v -> sprintf ";MaxAge=%s" (Int64.to_string v)) ~default in
    let domain =
      Cookie.get_domain c
      |> Option.value_map ~f:(sprintf ";Domain=%s") ~default in
    let path =
      Cookie.get_path c
      |> Option.value_map ~f:(sprintf ";Path=%s") ~default in
    let secure =
      if Cookie.is_secure c then ";Secure" else "" in
    let http_only =
      if Cookie.is_http_only c then ";HttpOnly" else "" in
    let same_site =
      Cookie.get_same_site c
      |> Option.value_map ~f:(fun v -> sprintf ";SameSite=%s" (Cookie.Same_site.show v)) ~default in
    String.concat
      [cookie_pair; expires; max_age; domain; path; secure; http_only; same_site ] in
  List.map cookies ~f:to_set_cookie_header

let find_cookie cookies ~key =
  List.find cookies ~f:(fun c -> c |> Cookie.get_key |> String.equal key)

let find_value cookies ~key =
  find_cookie cookies ~key |> Option.map ~f:Cookie.get_value

let set new_cookie cookies =
  cookies
  |> List.filter ~f:(fun c ->
      not (String.equal (Cookie.get_key c) (Cookie.get_key new_cookie)))
  |> List.cons new_cookie

