(** Types and functions for manipulation on HTTP headers. *)
module Headers : sig
  type t

  val show: t -> string

  (** Creates new empty headers object *)
  val empty: t

  (** Creates new headers object from a given list. *)
  val of_list: (string * string) list -> t

  (** Converts headers to a list *)
  val to_list: t -> (string * string) list

  (** Returns the first value of the header [~key] if it exists *)
  val find: t -> key:string -> string option

  (** Returns all values of the header [~key] if it exists *)
  val find_all: t -> key:string -> string list

  (** Sets the value of the header [~key] to [~value]*)
  val set: t -> key:string -> value:string -> t

  (** Sets the values of the header [~key] to [~values]*)
  val set_all: t -> key:string -> values:string list -> t

  (** Adds [~value] to the values of the header [~key] *)
  val add: t -> key:string -> value:string -> t

  (** Removes the header [~key] completely *)
  val remove: t -> key:string -> t
end

(** Types and functions representing body of HTTP request or response *)
module Body : sig
  type t

  val show: t -> string

  (** Creates new empty body *)
  val empty: t

  (** Creates new body from a given string *)
  val of_string: string -> t

  (** Creates new body from a given Lwt stream *)
  val of_stream: string Lwt_stream.t -> t

  (** Returns the whole body if it was given in a single chunk, e.g.
      constructed using [of_string] or [empty] functions. Returns [None] if the
      body is streamed, e.g. constructed using [of_stream] function or come from
      actual request/response with [Transfer-Encoding: chunked] header.

      Use [read_up_to] to read the streamed body or [to_stream] and [Lwt_stream]
      module functions to process it chunk by chunk.
  *)
  val to_string: t -> string option

  (** Returns Lwt stream containing the body *)
  val to_stream: t -> string Lwt_stream.t

  (** Reads up to [~max_bytes] bytes (default: 1MB) from the body.

      IMPORTANT: the actual size could differ from [~max_bytes] because body is
      read in chunks, not byte by byte.
  *)
  val read_up_to: ?max_bytes:int -> t -> string Lwt.t

  (** Returns the length of the body in bytes if it is known, e.g. the body is
      not streamed *)
  val length: t -> int option
end

(** Types and functions for accessing urlencoded form body *)
module Form : sig
  type t

  val show: t -> string

  (** Reads and decodes given body constructing a form from it *)
  val of_body: ?max_bytes:int -> Body.t -> t Lwt.t

  (** Returns all form params associated with [~key] *)
  val find_params: t -> key:string -> string list option

  (** Returns the first form param associated with [~key] *)
  val find_param: t -> key:string -> string option
end

(** Type and functions for representing a single cookie *)
module Cookie : sig
  module Same_site : sig
    type t = Strict | Lax | None
    val show: t -> string
  end

  type t

  val show: t -> string
  val equal: t -> t -> bool

  (** Creates new cookie from the given parameters. *)
  val make:
    ?expires:Core.Time.t ->
    ?max_age:int64 ->
    ?domain:string ->
    ?path:string ->
    ?secure:bool ->
    ?http_only:bool ->
    ?same_site:Same_site.t ->
    (string * string) ->
    t

  (** Converts cookie to a key-value pair *)
  val to_pair: t -> (string * string)

  (** Returns cookie key *)
  val get_key: t -> string

  (** Returns cookie value *)
  val get_value: t -> string

  (** Returns cookie 'expires' attribute value *)
  val get_expires: t -> Core.Time.t option

  (** Returns cookie 'max_age' attribute value *)
  val get_max_age: t -> int64 option

  (** Returns cookie 'domain' attribute value *)
  val get_domain: t -> string option

  (** Returns cookie 'path' attribute value *)
  val get_path: t -> string option

  (** Returns cookie 'secure' attribute value *)
  val is_secure: t -> bool

  (** Returns cookie 'http_only' attribute value *)
  val is_http_only: t -> bool

  (** Returns cookie 'same_site' attribute value *)
  val get_same_site: t -> Same_site.t option

  (** Updates cookie key *)
  val set_key: string -> t -> t

  (** Updates cookie value *)
  val set_value: string -> t -> t

  (** Updates cookie 'expires' attribute *)
  val set_expires: Core.Time.t option -> t -> t

  (** Updates cookie 'max_age' attribute *)
  val set_max_age: int64 option -> t -> t

  (** Updates cookie 'domain' attribute *)
  val set_domain: string option -> t -> t

  (** Updates cookie 'path' attribute *)
  val set_path: string option -> t -> t

  (** Updates cookie 'secure' attribute *)
  val set_secure: bool -> t -> t

  (** Updates cookie 'http_only' attribute *)
  val set_http_only: bool -> t -> t

  (** Updates cookie 'same_site' attribute *)
  val set_same_site: Same_site.t option -> t -> t
end

(** Types and functions for accessing and managing cookies *)
module Cookie_jar : sig
  type t

  val show: t -> string

  (** Creates new empty cookie jar *)
  val empty: t

  (** Constructs new cookie jar from the given list of cookies.
      In case of cookies with duplicate keys, only one of them is kept. *)
  val of_list: Cookie.t list -> t

  (** Parses cookies from the given 'Cookie' header value *)
  val of_cookie_header: string -> t

  (** Parses cookies from the given 'Set-Cookie' header values *)
  val of_set_cookie_headers: string list -> t

  (** Returns list of key-value pairs in this cookie jar *)
  val to_list: t -> Cookie.t list

  (** Serializes cookies to a string that could be used as a 'Cookie' header value *)
  val to_cookie_header: t -> string

  (** Serializes cookies to a list of strings that could be used as a
      'Set-Cookie' header values *)
  val to_set_cookie_headers: t -> string list

  (** Returns value of the cookie [~key] if it exists *)
  val find_value: t -> key:string -> string option

  (** Returns the whole cookie [~key] if it exists *)
  val find_cookie: t -> key:string -> Cookie.t option

  (** Adds or replaces existing cookie with the same key *)
  val set: Cookie.t -> t -> t
end

(** Types and functions for HTTP request manipilation *)
module Request : sig
  type t

  val show: t -> string

  (** Returns remote address of the client making the request or None if request
      is created manually. *)
  val get_client_addr: t -> Unix.sockaddr option

  (** Returns HTTP method in all caps, e.g. GET, POST, etc *)
  val get_http_method: t -> string

  (** Returns URI of the request *)
  val get_uri: t -> Uri.t

  (** Returns path part of the request URL as a list of chunks split by '/' *)
  val get_path: t -> string list

  (** Returns query string of the request URL as a list of key/values pairs *)
  val get_query_string: t -> (string * string list) list

  (** Returns headers of the request *)
  val get_headers: t -> Headers.t

  (** Returns body of the request *)
  val get_body: t -> Body.t

  (** Returns cookies present in the request 'Cookie' header *)
  val get_cookies: t -> Cookie_jar.t

  (** Returns all query string params associated with [~key] *)
  val find_query_params: t -> key:string -> string list option

  (** Returns the first query string param associated with [~key] *)
  val find_query_param: t -> key:string -> string option

  (** Returns the value of the header [~key] if it exists *)
  val find_header: t -> key:string -> string option

  (** Creates new HTTP request with a given [~http_method] and [~uri] *)
  val make: http_method:string -> uri:string -> t

  (** Creates new GET HTTP request for a given URI *)
  val get: string -> t

  (** Creates new POST HTTP request for a given URI *)
  val post: string -> t

  (** Overwrites the body of the request with a given one *)
  val with_body: Body.t -> t -> t

  (** Overwrites the body of the request with a given string *)
  val with_string_body: string -> t -> t

  (** Overwrites the list of headers with a given one *)
  val with_headers: Headers.t -> t -> t

  (** Overwrites the header of the request with a given value. *)
  val with_header: key:string -> value:string -> t -> t

  (** Overwrites value of the 'Cookie' header with the cookies from the given
      jar *)
  val with_cookies: Cookie_jar.t -> t -> t
end

(** Type and functions for HTTP response manipulation *)
module Response : sig
  type t

  val show: t -> string

  (** Returns status of the response *)
  val get_status: t -> int

  (** Returns headers of the response *)
  val get_headers: t -> Headers.t

  (** Returns body of the response *)
  val get_body: t -> Body.t

  (** Returns cookies present in the response 'Set-Cookie' header *)
  val get_cookies: t -> Cookie_jar.t

  (** Returns the value of the header [~key] if it exists *)
  val find_header: t -> key:string -> string option

  (** Returns an empty 200 response *)
  val ok: t

  (** Returns an empty response with a given status code *)
  val of_status: int -> t

  (** Returns a response redirecting to the given URL.
      If [?status] is not set, 303 is used.
      Response body is set to an HTML page containing a link to the given URL. *)
  val redirect: ?status:int -> string -> t

  (** Overwrites the body of the response with a given one. *)
  val with_body: Body.t -> t -> t

  (** Sets the body of the response to a given string. *)
  val with_string_body: string -> t -> t

  (** Overwrites the list of headers with a given one *)
  val with_headers: Headers.t -> t -> t

  (** Sets the header of the response to a given value. *)
  val with_header: key:string -> value:string -> t -> t

  (** Overwrites value of the 'Set-Cookie' header with the cookies from the
      given jar *)
  val with_cookies: Cookie_jar.t -> t -> t
end

(** Client-side functions *)
module Client : sig
  (** Issues an HTTP request using information provided in a given request
      object and returns a response from the server. *)
  val send: Request.t -> Response.t Lwt.t
end

(** Type of server-side handler that processes requests *)
module Handler : sig
  type t = Request.t -> Response.t Lwt.t
end

(** Type of middleware that checks and transforms requests and passes them down
    the chain *)
module Middleware : sig
  type t = Handler.t -> Handler.t

  (** Request/response logging middleware *)
  module Log : sig
    (** Logs every request and response in Common Log Format using [~log]
        function *)
    val middleware: log:(string -> unit) -> t
  end
end

(** Server-side functions *)
module Server : sig
  (** Starts a new HTTP server listening on given [~address] and [~port].
      Each request is processed with a given handler. Optional [?on_exn]
      parameter is a function called whenever an exception is raised inside the
      handler (default is to log the exception to stderr). *)
  val listen: ?on_exn:(exn -> unit) -> address:string -> port:int -> Handler.t -> unit Lwt.t
end
