open Base
open Okweb

let test_to_cookie_header =
  let open Alcotest in
  test_case "to_cookie_header serialization" `Quick begin fun () ->
    let cookies = Cookie_jar.of_list [
        Cookie.make ("bar", "bar_value");
        Cookie.make ("foo", "foo value");
      ] in
    check string
      "checking serialized value"
      "bar=bar_value;foo=foo%20value"
      (Cookie_jar.to_cookie_header cookies);
  end

let test_of_cookie_header =
  let open Alcotest in
  test_case "of_cookie_header deserialization" `Quick begin fun () ->
    let cookies = Cookie_jar.of_cookie_header "foo=\"foo%20value\" ; bar = bar_value  " in
    check (slist (pair string string) (fun (k1, _) (k2, _) -> String.compare k1 k2))
      "checking deserialized value"
      ["foo", "foo value"; "bar", "bar_value"]
      (cookies |> Cookie_jar.to_list |> List.map ~f:Cookie.to_pair);
  end

let test_to_set_cookie_header_expires_attr =
  Alcotest.test_case "to_set_cookie_header expires attr format serialization" `Quick begin fun () ->
    let expires = Core.Time.of_span_since_epoch (Core.Time.Span.of_int_sec 1586675796) in
    let cookies = Cookie_jar.of_list [Cookie.make ~expires ("key", "value")] in
    Alcotest.(check (list string))
      "checking serialized cookie with expires attribute"
      ["key=value;Expires=Sun, 12 Apr 2020 07:16:36 GMT"]
      (Cookie_jar.to_set_cookie_headers cookies);
  end

let cookie_testable =
  Alcotest.testable Fmt.(using Cookie.show string) Cookie.equal

let test_of_set_cookie_headers_skips_invalid_attrs =
  let open Alcotest in
  test_case "of_set_cookie_headers skips invalid attributes" `Quick begin fun () ->
    let invalid_header =
      "key=value;Expires=invalid_value;MaxAge=invalid_value;Invalid_attr;HttpOnly;SameSite=whatever;Invalid_attr=Invalid_value" in
    check (list cookie_testable)
      "checking deserialized cookie with invalid attributes"
      [Cookie.make ~http_only:true ("key", "value")]
      Cookie_jar.(of_set_cookie_headers [invalid_header] |> to_list);
  end

let cookie_gen =
  let open QCheck.Gen in
  let is_separator c =
    Char.is_whitespace c || String.contains "()<>@,;:\\\"/[]?={}" c in
  let key_alphabet =
    Char.all |> List.filter ~f:(fun c -> Char.is_print c && not (is_separator c)) in
  let attr_alphabet =
    Char.all |> List.filter ~f:(fun c -> Char.(is_print c && not (is_whitespace c) && c <> ';')) in
  (fun key value expires max_age domain path secure http_only same_site ->
     Cookie.make ?expires ?max_age ?domain ?path ?secure ?http_only ?same_site (key, value))
  <$> (small_string ~gen:(oneofl key_alphabet))
  <*> small_string
  <*> (opt ((fun s -> Core.Time.of_span_since_epoch (Core.Time.Span.of_int32_seconds s)) <$> ui32))
  <*> (opt ui64)
  <*> (opt (small_string ~gen:(oneofl attr_alphabet)))
  <*> (opt (small_string ~gen:(oneofl attr_alphabet)))
  <*> (opt bool)
  <*> (opt bool)
  <*> (opt (oneofl Cookie.Same_site.([Strict; Lax; None]))) [@warning "-45"]

let cookie_arbitrary =
  QCheck.make cookie_gen ~print:Cookie.show

let test_cookie_header_roundtrip =
  QCheck.Test.make QCheck.(list cookie_arbitrary)
    ~name:"cookie_header roundtrip (de)serialization"
    begin fun in_cookies ->
      let out_cookies = Cookie_jar.(
          in_cookies
          |> of_list
          |> to_cookie_header
          |> of_cookie_header
          |> to_list) in
      let cookie_kv_equal c1 c2 =
        String.equal (Cookie.get_key c1) (Cookie.get_key c2) &&
        String.equal (Cookie.get_value c1) (Cookie.get_value c2) in
      let compare c1 c2 = String.compare (Cookie.get_key c1) (Cookie.get_key c2) in
      List.equal cookie_kv_equal (List.dedup_and_sort ~compare in_cookies) (List.sort ~compare out_cookies)
    end

let test_set_cookie_headers_roundtrip =
  QCheck.Test.make QCheck.(list cookie_arbitrary)
    ~name:"set_cookie_headers roundtrip (de)serialization"
    begin fun in_cookies ->
      let out_cookies = Cookie_jar.(
          in_cookies
          |> of_list
          |> to_set_cookie_headers
          |> of_set_cookie_headers
          |> to_list) in
      let compare c1 c2 = String.compare (Cookie.get_key c1) (Cookie.get_key c2) in
      List.equal Cookie.equal (List.dedup_and_sort ~compare in_cookies) (List.sort ~compare out_cookies)
    end

let test_set_adds_or_replaces_cookie =
  let open Alcotest in
  test_case "set adds or replaces cookie" `Quick begin fun () ->
    let foo_1 = Cookie.make ("foo", "foo_1") in
    let foo_2 = Cookie.make ("foo", "foo_2") in
    let bar_1 = Cookie.make ("bar", "bar_1") in
    let cookies = Cookie_jar.(empty |> set foo_1 |> set bar_1 |> set foo_2) in
    check (option string)
      "checking 'foo' cookie value"
      (Some "foo_2")
      (Cookie_jar.find_value cookies ~key:"foo");
    check (option string)
      "checking 'bar' cookie value"
      (Some "bar_1")
      (Cookie_jar.find_value cookies ~key:"bar");
  end

let cases = List.concat [[
    test_to_cookie_header;
    test_of_cookie_header;
    test_to_set_cookie_header_expires_attr;
    test_set_adds_or_replaces_cookie;
    test_of_set_cookie_headers_skips_invalid_attrs;
  ]; List.map ~f:QCheck_alcotest.to_alcotest [
    test_cookie_header_roundtrip;
    test_set_cookie_headers_roundtrip;
  ]]
