open Alcotest
open Base
open Okweb

let test_create_headers_from_list =
  test_case "create headers from list" `Quick begin fun () ->
    let header_list = [
      "TEST-HEADER", "Test Value 1";
      "TEST-header", "Test Value 2";
      "test-header", "Test Value 3";
    ] in
    let headers = Headers.of_list header_list in
    check (list (pair string string))
      "headers preserve multiple values for the same key"
      header_list
      (Headers.to_list headers)
  end

let test_find_regardless_of_key_case =
  test_case "find regardless of key's case" `Quick begin fun () ->
    let value = "Test Value" in
    let headers = Headers.of_list ["test-header", value] in
    let cases = [
      "test-header";
      "TEST-HEADER";
      "test-HEADER";
      "tEsT-hEaDeR";
    ] in
    List.iter cases ~f:begin fun key ->
      check (option string)
        (Printf.sprintf "searching value for the header '%s'" key)
        (Some value)
        (Headers.find headers ~key)
    end
  end


let test_find_all_regardless_of_key_case =
  test_case "find_all regardless of key's case" `Quick begin fun () ->
    let headers = Headers.of_list [
        "test-header", "Test Value 1";
        "TEST-HEADER", "Test Value 2";
      ] in
    let cases = [
      "test-header";
      "TEST-HEADER";
      "test-HEADER";
      "tEsT-hEaDeR";
    ] in
    List.iter cases ~f:begin fun key ->
      check (list string)
        (Printf.sprintf "searching all values for the header '%s'" key)
        ["Test Value 1"; "Test Value 2"]
        (Headers.find_all headers ~key)
    end
  end

let test_remove_header =
  test_case "remove regardless of key's case" `Quick begin fun () ->
    let headers = Headers.of_list [
        "header-1", "value-1";
        "header-2", "value-2";
        "HEADER-1", "value-3";
      ] in
    check (list string)
      "checking values of header-1 (before removal)"
      ["value-1"; "value-3"]
      (Headers.find_all headers ~key:"header-1");
    check (list string)
      "checking values of header-2 (before removal)"
      ["value-2"]
      (Headers.find_all headers ~key:"header-2");

    let headers = Headers.remove headers ~key:"HEAder-1" in

    check (list string)
      "checking values of header-1 (after removal)"
      []
      (Headers.find_all headers ~key:"header-1");
    check (list string)
      "checking values of header-2 (after removal)"
      ["value-2"]
      (Headers.find_all headers ~key:"header-2");
  end

let test_add_header_appends_value =
  test_case "add header appends value" `Quick begin fun () ->
    let headers = Headers.of_list ["header-1", "value-1"] in
    let new_headers = Headers.add headers ~key:"header-1" ~value:"value-2" in
    check (list string)
      "checking values of header-1"
      ["value-1"; "value-2"]
      (Headers.find_all new_headers ~key:"header-1");
  end

let test_add_header_creates_value =
  test_case "add header creates value" `Quick begin fun () ->
    let headers = Headers.of_list ["header-1", "value-1"] in
    let new_headers = Headers.add headers ~key:"header-2" ~value:"value-2" in
    check (list string)
      "checking values of header-1"
      ["value-1"]
      (Headers.find_all new_headers ~key:"header-1");
    check (list string)
      "checking values of header-2"
      ["value-2"]
      (Headers.find_all new_headers ~key:"header-2");
  end

let cases = [
  test_create_headers_from_list;
  test_find_regardless_of_key_case;
  test_find_all_regardless_of_key_case;
  test_remove_header;
  test_add_header_appends_value;
  test_add_header_creates_value;
]
